#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 17 13:34:59 2019

@author: atekawade
"""

import numpy as np







def get_patches(img, patch_size = None, steps = None):
    
    stepy, stepx = steps
    my, mx = img.shape
    py, px = patch_size[0], patch_size[1]
    nx, ny = int(np.ceil(mx/px)), int(np.ceil(my/py))
    
    img = np.asarray([img[ii*stepy:ii*stepy+py] for ii in range(ny)])
    img = np.asarray([[img[jj,:,ii*stepx:ii*stepx+px] for ii in range(nx)] for jj in range(img.shape[0])])
    
    return img


def get_stepsize(img_shape, patch_size):
    
    # Find optimum number of patches to cover full image
    my, mx = img_shape
    py, px = patch_size
    nx, ny = int(np.ceil(mx/px)), int(np.ceil(my/py))
    stepx = (mx-px) // (nx-1) if mx != px else 0
    stepy = (my-py) // (ny-1) if my != py else 0
    
    return (stepy, stepx)



def recon_patches(img, img_shape = None, steps = None):
    
    if img.ndim != 4:
        raise ValueError("Input must be 4D array.")
    
    ny, nx, py, px = img.shape
    stepy, stepx = steps
    
    
    new_img = np.zeros((img_shape))
    
    for ii in range(ny):
        for jj in range(nx):
            new_img[ii*stepy:ii*stepy+py,jj*stepx:jj*stepx+px] = img[ii,jj]
        
    return new_img


def calc_resdrop(img_shape, patch_size, n_max = 3):
    
    y_orig, x_orig, = img_shape
    
    yres = 1
    y_new = y_orig
    while y_new > patch_size[0]*n_max:
        yres += 1
        y_new =  int(np.ceil(y_orig/yres))
    
    xres = 1
    x_new = x_orig
    while x_new > patch_size[1]*n_max:
        xres += 1
        x_new =  int(np.ceil(x_orig/xres))
    
    return yres, xres #, y_new, x_new










def ssrecon_patches(img, img_shape = None, steps = None):
    
    ny, nx = img.shape[:2]
    stepy, stepx = steps
    p = img.shape[-1]
    
    new_img = np.zeros((img_shape))
    
    for ii in range(ny):
        for jj in range(nx):
            new_img[ii*stepy:ii*stepy+p,jj*stepx:jj*stepx+p] = img[ii,jj]
        
    return new_img















#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
A module for estimating
    1. signal-to-noise ratio (SNR) for binarizable datasets.  
    2. accuracy metrics for segmentation maps.  

"""

import numpy as np
from multiprocessing import cpu_count
import matplotlib as mpl
from ct_segnet.data_utils.data_io import Parallelize

import matplotlib.pyplot as plt


def calc_SNR(img, seg_img, labels = (0,1), mask_ratio = None):
    """
    SNR =  1     /  s*sqrt(std0^^2 + std1^^2)  
    where s = 1 / (mu1 - mu0)  
    mu1, std1 and mu0, std0 are the mean / std values for each of the segmented regions respectively (pix value = 1) and (pix value = 0).  
    seg_img is used as mask to determine stats in each region.  

    Parameters
    ----------
    img : np.array  
        raw input image (2D or 3D)  
    
    seg_img : np.array  
        segmentation map (2D or 3D)  
        
    labels : tuple  
        an ordered list of two label values in the image. The high value is interpreted as the signal and low value is the background.  
        
    mask_ratio : float or None
        If not None, a float in (0,1). The data are cropped such that the voxels / pixels outside the circular mask are ignored.  

    Returns
    -------
    float
        SNR of img w.r.t seg_img  

    """
    
    # handle circular mask  
    if mask_ratio is not None:
        crop_val = int(img.shape[-1]*0.5*(1 - mask_ratio/np.sqrt(2)))
        crop_slice = slice(crop_val, -crop_val)    

        if img.ndim == 2: # 2D image
            img = img[crop_slice, crop_slice]
            seg_img = seg_img[crop_slice, crop_slice]
        elif img.ndim == 3: # 3D image
            vcrop = int(img.shape[0]*(1-mask_ratio))
            vcrop_slice = slice(vcrop, -vcrop)
            img = img[vcrop_slice, crop_slice, crop_slice]
            seg_img = seg_img[vcrop_slice, crop_slice, crop_slice]
            
        
    pix_1 = img[seg_img == labels[1]]
    pix_0 = img[seg_img == labels[0]]
    mu1 = np.mean(pix_1)
    mu0 = np.mean(pix_0)
    s = abs(1/(mu1 - mu0))
    std1 = np.std(pix_1)
    std0 = np.std(pix_0)
    std = np.sqrt(0.5*(std1**2 + std0**2))
    std = s*std
    return 1/std


def ROC(thresh, true_img = None, seg_img = None):
    """
    Receiver Operating Characteristics (ROC) curve  
    
    Parameters
    ----------
    thresh : float
            threshold value
            
    true_img : numpy.array
            ground truth segmentation map (ny, nx)
            
    seg_img : numpy.array
            predicted segmentation map (ny, nx)
    
    Returns
    -------
    tuple
        FPR, TPR        
            
    """
    y_p = np.zeros_like(seg_img)
    y_p[seg_img > thresh] = 1
    true_img = np.copy(true_img)

    TN = np.sum((1-true_img)*(1-y_p)).astype(np.float32)
    
    FP = np.sum((1-true_img)*y_p).astype(np.float32)
    
    TNR = TN / (TN + FP)
    FPR = 1 - TNR
    
    TP = np.sum(true_img*y_p).astype(np.float32)
    
    FN = np.sum(true_img*(1-y_p)).astype(np.float32)
    
    TPR = TP / (TP +  FN)
    
    return (FPR, TPR)

def calc_jac_acc(true_img, seg_img):
    """
    Parameters
    ----------
    true_img : np.array
            ground truth segmentation map (ny, nx)
            
    seg_img : np.array
            predicted segmentation map (ny, nx)
     
    Returns
    -------
    float
        Jaccard accuracy or Intersection over Union  
    """
    seg_img = np.round(np.copy(seg_img))
    
    jac_acc = (np.sum(seg_img*(true_img == 1)) + 1) / (np.sum(seg_img) + np.sum((true_img == 1)) - np.sum(seg_img*(true_img == 1)) + 1)
    return jac_acc

def calc_dice_coeff(true_img, seg_img):
    """
    Parameters
    ----------
    true_img : np.array
            ground truth segmentation map (ny, nx)
            
    seg_img : np.array
            predicted segmentation map (ny, nx)
     
    Returns
    -------
    float
        Dice coefficient  

    """
    seg_img = np.round(np.copy(seg_img))
    
    dice = (2*np.sum(seg_img*(true_img == 1)) + 1) / (np.sum(seg_img) + np.sum((true_img == 1)) + 1)
    return dice

def fidelity(true_imgs, seg_imgs, tolerance = 0.95):
    """
    Fidelity is number of images with IoU > tolerance  
    
    Parameters
    ----------
    tolerance : float
                tolerance (default  = 0.95)
                
    true_imgs : numpy.array
                list of ground truth segmentation maps (nimgs, ny, nx)
                
    seg_imgs  : numpy.array
                list of predicted segmentation maps (nimgs, ny, nx)
     
    Returns
    -------
    float
        Fidelity  
    """

    XY = [(true_imgs[ii], seg_imgs[ii]) for ii in range(true_imgs.shape[0])]
    del true_imgs
    del seg_imgs
    
    jac_acc = np.asarray(Parallelize(XY, calc_jac_acc, procs = cpu_count()))
    
    mean_IoU = np.mean(jac_acc)

    jac_fid = np.zeros_like(jac_acc)
    jac_fid[jac_acc > tolerance] = 1
    jac_fid = np.sum(jac_fid).astype(np.float32) / np.size(jac_acc)
    
    
    return jac_fid, mean_IoU, jac_acc









# -*- coding: utf-8 -*-
"""

"""
import os
import numpy as np
import matplotlib as mpl
#mpl.use('Agg')
import matplotlib.pyplot as plt
import h5py
import sys
from tensorflow import keras
from IPython.display import clear_output
import random
import pandas as pd
from multiprocessing import cpu_count
from ct_segnet.data_utils.data_io import Parallelize
from ct_segnet.stats import *#ROC, calc_jac_acc, calc_dice_coeff, fidelity
from ct_segnet.data_utils.data_augmenter import run_augmenter

def _norm(y):
    if (y.max() != 1.0) | (y.min() != 0.0):
        y = (y - np.min(y)) / (np.max(y) - np.min(y))
    return y.astype(np.uint8)

def data_generator(X, Y, batch_size, \
                   check_norm = False, \
                   augmenter_todo = None, \
                   min_SNR = 2.0, \
                   inplace = False, nprocs = 1):
    """
    Generator that yields randomly sampled data pairs of size batch_size. X, Y are DataFile instance pairs of train / test / validation data.  
    
    Parameters
    ----------
    X: DataFile
        Input images  
    
    Y: DataFile
        Ground truth segmentation  
        
    check_norm : bool
        Ensure that segmentation labels are 0 and 1 only. Should be False if your ground truth data contains ignored pixels.  
        
    augmenter_todo  : list
        list of strings from "rotate", "flip", "gaussian noise", "invert intensity"
    
    min_SNR : float
        minimum allowable signal-to-noise ratio (SNR)
    
    Returns
    -------
    tuple
        X, Y numpy arrays each of shape (n_batch, ny, ny)  
    
    """
    while True:
        idxs = sorted(random.sample(range(X.d_shape[0]), batch_size))
        x = X.read_sequence(idxs)
        y = Y.read_sequence(idxs)
        
        if check_norm:
            y = _norm(y)
        
        if augmenter_todo is not None:
            x, y = run_augmenter(x, y, \
                                 min_SNR = min_SNR, \
                                 to_do = augmenter_todo, \
                                 inplace = inplace, \
                                 nprocs = nprocs)

        yield (x[...,np.newaxis], y[...,np.newaxis])

class Logger(keras.callbacks.Callback):
    """  
    An instance of Logger can be passed to keras model.fit to log stuff at epochs.  
    
    Parameters
    ----------
    model_paths : dict
        with keys = ["name", "history", "file"]  
    
    Xtest : DataFile
        input images from test data for calculating model accuracy  
    
    Ytest : DataFile
        ground truth segmentation map from test data  
    
    df_prev : pandas.DataFrame
        If retraining, pass previous epochs data  
    
    N : int
        autosave frequency (in epochs)  
    
    n_test : int
        number of test image pairs to be sampled every epoch  
    
    """
    def __init__(self, Xtest, Ytest, model_paths, N,\
                 df_prev = None, n_test = None, \
                 check_norm = False, augmenter_todo = None,):
        
        self.test_dg = data_generator(Xtest, Ytest, n_test, \
                                      check_norm = check_norm, \
                                      augmenter_todo = augmenter_todo)
        self.N = N
        self.df_prev = df_prev
        self.model_name = model_paths['name']
        self.model_history = model_paths['history']
        self.model_file = model_paths['file']
            
    
    def on_train_begin(self, logs={}):
        
        if self.df_prev is not None:
            next_epoch = self.df_prev['epoch'].tolist()[-1] + 1
            print("\n\nRestarting training from epoch # %i..."%next_epoch)

        # INITIALIZE VECTORS
        self.i = 1 if self.df_prev is None else next_epoch
        self.x = [] if self.df_prev is None else self.df_prev['epoch'].tolist()
        self.losses = [] if self.df_prev is None else self.df_prev['loss'].tolist()
        self.val_losses = [] if self.df_prev is None else self.df_prev['val_loss'].tolist()
        self.test_losses = [] if self.df_prev is None else self.df_prev['test_loss'].tolist()
        self.metric = [] if self.df_prev is None else self.df_prev['metric'].tolist()
        self.acc_zeros = [] if self.df_prev is None else self.df_prev['acc_zeros'].tolist()
        self.acc_ones = [] if self.df_prev is None else self.df_prev['acc_ones'].tolist()

        self.logs = []

    def on_epoch_end(self, epoch, logs={}):
        
        self.fig, self.ax = plt.subplots(1,2, figsize = (12,5))
        self.ax[0].set_xlabel('epoch')
        self.ax[0].set_ylabel('loss')
        self.ax[1].set_xlabel('epoch')
        self.ax[1].set_ylabel('metrics')
        

        # APPEND DATA
        self.logs.append(logs)
        self.x.append(self.i)
        self.losses.append(logs.get('loss'))
        self.val_losses.append(logs.get('val_loss'))
        x_test, y_test = next(self.test_dg)
        evaluation = self.model.evaluate(x = x_test, y = y_test, verbose = 0)
        if type(evaluation) is list:
            if len(evaluation) == 4:
                self.test_losses.append(evaluation[0])
                self.metric.append(evaluation[1])
                self.acc_zeros.append(evaluation[2])
                self.acc_ones.append(evaluation[3])
            else:
                raise ValueError("\n number of metrics provided does not match plotter's configuration.")
        else:
            raise ValueError("\n expecting metric definitions but none found.")
        
        print("val_loss on test data: %.4f"%self.test_losses[-1])
        print("metric evaluated on test data: %.4f"%self.metric[-1])
        print("acc_zeros: %.2f, acc_ones: %.2f"%(self.acc_zeros[-1], self.acc_ones[-1]))
        
        clear_output(wait=True)
        self.ax[0].plot(self.x, self.losses, label="loss")
        self.ax[0].plot(self.x, self.val_losses, label="val_loss")
        self.ax[0].plot(self.x, self.test_losses, label = "test_data val_loss")
        self.ax[0].legend(loc = 'upper right')
        
        
        self.ax[1].plot(self.x, self.acc_zeros, label = 'acc_zeros')
        self.ax[1].plot(self.x, self.acc_ones, label = 'acc_ones')
        self.ax[1].plot(self.x, self.metric, '--k', label = "IoU")
        self.ax[1].legend(loc = 'upper right')
        
#         plt.show()
        plt.savefig(os.path.join(self.model_history, 'Loss_plot_' + self.model_name + '.png'))
        plt.close()     
        
        if self.i % self.N == 0:

            # Save loss data in .csv file
            df = pd.DataFrame()
            df['epoch'] = self.x
            df['loss'] = self.losses
            df['val_loss'] = self.val_losses
            df['test_loss'] = self.test_losses
            df['metric'] = self.metric
            df['acc_zeros'] = self.acc_zeros
            df['acc_ones'] = self.acc_ones
            df.set_index('epoch')
            df.to_csv(os.path.join(self.model_history,self.model_name+".csv"))

        
            if self.test_losses[-1] < 1.5*self.test_losses[-self.N]:
                print("Saving model at epoch # %i..."%self.i)
                self.model.save(self.model_file)
                self.model.save_weights(os.path.join(self.model_history, 'weights_' + self.model_name + '.h5'))
            else:
                print("Terminate Training: Overfitting detected from increasing test loss...")
                # self.model.stop_training = True
        
        self.i += 1


    def on_train_end(self, logs = {}):
        
        self.logs.append(logs)
        
        

def save_datasnaps(dg, model_history, n_imgs = 20):
    """
    """
    if not os.path.exists(os.path.join(model_history,"data_snaps")):
        os.makedirs(os.path.join(model_history,"data_snaps"))
        
        
    # Load and save random train data images
    x, y = next(dg)
    x, y = x[...,0], y[...,0]
    for jj in x.shape[0]:
        fig, ax = plt.subplots(1,2, figsize = (6,3))
        ax[0].axis('off')
        ax[0].imshow(y[jj])
        ax[1].axis('off')
        ax[1].imshow(x[jj])
        fig.suptitle("MIN: %.2f, MAX: %.2f, MEAN: %.2f"%(np.min(x[jj]), np.max(x[jj]), np.mean(x[jj])))
        plt.savefig(os.path.join(model_history,"data_snaps",'randomdata_%5d.png'%jj))
        plt.close()


    return






    
    

def save_results(dg, model_results, segmenter):
    """Save some results on test images into a folder in the path to model repo
    """
    x_test, y_test = next(dg)
    y_pred = segmenter.predict(x_test)
    y_pred = np.round(y_pred)
    
    x_test, y_test, y_pred = x_test[...,0], y_test[...,0], y_pred[...,0]
    
    if not os.path.exists(os.path.join(model_results,"data_snaps")):
        os.makedirs(os.path.join(model_results,"data_snaps"))
    
    for ii in range(x_test.shape[0]):
            
            fig, ax = plt.subplots(1,3, figsize = (9,3))
            ax[0].axis('off')
            ax[0].imshow(x_test[ii])
            ax[1].axis('off')
            ax[1].imshow(y_test[ii])
            ax[2].axis('off')
            ax[2].imshow(y_pred[ii])
            
            #jac_acc = (np.sum(y_pred[ii]*y_test[ii]) + 1) / (np.sum(y_pred[ii]) + np.sum(y_test[ii]) - np.sum(y_pred[ii]*y_test[ii]) + 1)
            test_loss, jac_acc, acc_zeros, acc_ones = segmenter.evaluate(x = x_test[ii][np.newaxis,...,np.newaxis], y = y_test[ii][np.newaxis,...,np.newaxis], verbose = 0)
            
            fig.suptitle("IMG_MEAN: %.2f, JAC_ACC: %.2f, LOSS: %.2f"%(np.mean(x_test[ii]),jac_acc, test_loss))
            plt.savefig(os.path.join(model_results,"data_snaps",'snap_%05d.png'%ii))
            plt.close()

    return





















#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module defines some custom loss functions and metrics that are used to train and evaluate a U-net-like tf.Keras.model. This require the input to be a tensor.

Note: Some of these metrics are implemented in ct_segnet.stats to receive numpy.array as inputs.  

"""


import tensorflow as tf
from tensorflow.keras import backend as K


# Parameters for weighted cross-entropy and focal loss - alpha is higher than 0.5 to emphasize loss in "ones" or metal pixels.
eps = 1e-12
alpha = 0.75
gamma = 2.0

def IoU(y_true, y_pred):
    """
    :return: intersection over union accuracy
    
    Parameters
    ----------
    
    y_true  : tensor
            Ground truth tensor of shape (batch_size, n_rows, n_cols, n_channels)
    y_pred  : tensor
            Predicted tensor of shape (batch_size, n_rows, n_cols, n_channels)
    
    """
#     this is an old implementation that does not assume ignored pixels
#     y_pred = K.round(y_pred)
#     intersection = K.sum(y_true*y_pred)
#     union = K.sum(y_pred) + K.sum(y_true) - intersection
#     acc = (intersection + 1.) / (union + 1.)

#     this implementation assumes ignored pixel labels > 1 in y_true
    y_pred = K.round(y_pred)
    y_pred = tf.cast(y_pred, tf.int32)
    y_true = tf.cast(y_true, tf.int32)
    intersection = K.sum(tf.where(tf.equal(y_true,1), y_true*y_pred, 0))
    union = K.sum(y_pred) + K.sum(tf.where(tf.equal(y_true,1), 1, 0)) - intersection
    intersection = tf.cast(intersection, tf.float32)
    union = tf.cast(union, tf.float32)
    acc = (intersection + 1.) / (union + 1.)
    
    return acc



def acc_zeros(y_true, y_pred):
    """
    :return: accuracy in predicting zero values = TN/(TN + FP)
    
    Parameters
    ----------
    
    y_true  : tensor
            Ground truth tensor of shape (batch_size, n_rows, n_cols, n_channels)
    y_pred  : tensor
            Predicted tensor of shape (batch_size, n_rows, n_cols, n_channels)
    
    """
    
    # Define accuracy of zeros
    y_pred = K.round(y_pred)
    y_pred = tf.cast(y_pred, tf.int32)
    y_true = tf.cast(y_true, tf.int32)

    
#     true_negatives = K.sum((1-y_true)*(1-y_pred))
#     false_positives = K.sum((1-y_true)*y_pred)
#     p = (true_negatives + eps) / (true_negatives + false_positives + eps)

#     This version assumes ignored pixels
    true_negatives = K.sum(tf.where(tf.equal(y_true,0), 1, 0) * (1-y_pred))
    false_positives = K.sum(tf.where(tf.equal(y_true,0), 1 ,0) * y_pred)
    true_negatives = tf.cast(true_negatives, tf.float32)
    false_positives = tf.cast(false_positives, tf.float32)
    p = (true_negatives + eps) / (true_negatives + false_positives + eps)
    
    return p

def acc_ones(y_true, y_pred):
    """
    :return: accuracy in predicting ones = TP/(TP + FN)
    
    Parameters
    ----------
    
    y_true  : tensor
            Ground truth tensor of shape (batch_size, n_rows, n_cols, n_channels)
    y_pred  : tensor
            Predicted tensor of shape (batch_size, n_rows, n_cols, n_channels)
    
    """
    
    # Define accuracy of ones
    y_pred = K.round(y_pred)
    y_pred = tf.cast(y_pred, tf.int32)
    y_true = tf.cast(y_true, tf.int32)

    
    true_positives = K.sum(tf.where(tf.equal(y_true,1), 1, 0)*y_pred)
    false_negatives = K.sum(    tf.where(tf.equal(y_true,1), 1, 0)       * (1-y_pred))
    true_positives = tf.cast(true_positives, tf.float32)
    false_negatives = tf.cast(false_negatives, tf.float32)
    r = (true_positives + eps) / (true_positives + false_negatives + eps)
    
    return r


def _binary_lossmap(y_true, y_pred):
    # y_true, y_pred are tensors of shape (batch_size, img_h, img_w, n_channels)
    pt_1 = tf.where(tf.equal(y_true, 1), y_pred, tf.ones_like(y_pred))
    pt_0 = tf.where(tf.equal(y_true, 0), y_pred, tf.zeros_like(y_pred))
    return pt_1, pt_0

def my_binary_crossentropy(y_true, y_pred):
    """
    :return: loss value
    
    This is my own implementation of binary cross-entropy. Nothing special.
    
    Parameters
    ----------
    
    y_true  : tensor
            Ground truth tensor of shape (batch_size, n_rows, n_cols, n_channels)
    y_pred  : tensor
            Predicted tensor of shape (batch_size, n_rows, n_cols, n_channels)
    
    """

    pt_1, pt_0 = _binary_lossmap(y_true, y_pred)
    loss_map = -K.log(pt_1 + eps)-K.log(1. - pt_0 + eps)
    return tf.reduce_mean(loss_map)

def focal_loss(y_true, y_pred):
    """
    :return: loss value
    
    Focal loss is defined here: https://arxiv.org/abs/1708.02002
    Using this provides improved fidelity in unbalanced datasets: 
    Tekawade et al. https://doi.org/10.1117/12.2540442
    
    
    Parameters
    ----------
    
    y_true  : tensor
            Ground truth tensor of shape (batch_size, n_rows, n_cols, n_channels)
    y_pred  : tensor
            Predicted tensor of shape (batch_size, n_rows, n_cols, n_channels)
    
    """

    pt_1, pt_0 = _binary_lossmap(y_true, y_pred)
    loss_map = -alpha*K.log(pt_1 + eps)*K.pow(1. - pt_1,gamma) - (1-alpha)*K.log(1. - pt_0 + eps)*K.pow(pt_0,gamma)
    return tf.reduce_mean(loss_map)


def weighted_crossentropy(y_true, y_pred):
    """
    :return: loss value
    
    Weighted cross-entropy allows prioritizing accuracy in a certain class (either 1s or 0s).
    
    Parameters
    ----------
    
    y_true  : tensor
            Ground truth tensor of shape (batch_size, n_rows, n_cols, n_channels)
    y_pred  : tensor
            Predicted tensor of shape (batch_size, n_rows, n_cols, n_channels)
    
    """

    pt_1, pt_0 = _binary_lossmap(y_true, y_pred)
    loss_map = -alpha*K.log(pt_1 + eps)-(1-alpha)*K.log(1. - pt_0 + eps)
    loss_map *= 2.0
    return tf.reduce_mean(loss_map)


def stdize_img(img):
    
    eps = tf.constant(1e-12, dtype = 'float32')
    #mean = tf.reduce_mean(img)
    #img = img - mean
    
    max_ = tf.reduce_max(img)
    min_ = tf.reduce_min(img)

    img = (img - min_ )  / (max_ - min_ + eps)
    return img

def standardize(imgs):
    
    return tf.map_fn(stdize_img, imgs)
    
objects = [IoU, acc_zeros, acc_ones, focal_loss, my_binary_crossentropy, weighted_crossentropy, stdize_img, standardize]

custom_objects_dict = {'tf': tf}
for item in objects:
    custom_objects_dict[item.__name__] = item



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
A memory-efficient interface to slice, read and write CT data. Tiff series and hdf5 data formats are currently supported.
"""
import os
import numpy as np
import pandas as pd # line 13 empty for good luck

import shutil
import h5py
import glob

import matplotlib.pyplot as plt
import matplotlib as mpl

import functools
from multiprocessing import cpu_count, Pool
from skimage.io import imread
from tifffile import imsave
from configargparse import ArgumentTypeError
import ast


DEBUG_MODE = True

def _message(_str, bool_in):
    """
    :meta private:
    """
    
    if bool_in:
        print(_str)
        
def handle_YN(_str):
    """
    :meta private:
    """
    _inp = input(_str)
    return str_to_bool(_inp)


def str_to_bool(_inp):
    """
    
    :meta private:
    
    """
    _inp = str(_inp)
    if _inp in ("yes", "Yes", "Y", "y", "True", "TRUE", "true"):
        return True
    elif _inp in ("no", "No", "N", "n", "False", "FALSE", "false"):
        return False
    else:
        raise ArgumentTypeError("Input not understood")
    return

def n_patches_type(s):
    """
    :meta private:
    """
    s = s.split('x')
    s = ','.join(s)
    return ast.literal_eval(s)

def crops_type(s):
    """
    :meta private:
    """
    s = s.split(':')
    s = ','.join(s)
    return ast.literal_eval(s)


class InputError(Exception):
    """
    :meta private:
    """
    def __init__(self, message):
        self.message = message


class DataFile():
    """An instance of a DataFile class points to a 3D dataset in a tiff sequence or hdf5 file. The interface includes read/write methods to retrieve the data in several ways (slices, chunks, down-sampled data, etc.)  
    
    For setting chunk size in hdf5, either chunk_shape > chunk_size > chunked_slice_size can be input. If two or more are provided, this order is used to select one.  
    
    Parameters
    ----------
    fname : str
        path to hdf5 filename or folder containing tiff sequence  

    tiff : bool
        True if fname is path to tiff sequence, else False  
    
    data_tag: str
        dataset name / path in hdf5 file. None if tiff sequence  
    
    VERBOSITY : int
        0 - print nothing, 1 - important stuff, 2 - print everything  
    
    d_shape : tuple
        shape of dataset; required for non-existent dataset only  
    
    d_type : numpy.dtype
        data type for voxel data; required for non-existent dataset only  
    
    chunk_size: float
        in GB - size of a hyperslab of shape proportional to data shape    
    
    chunked_slice_size : float
        in GB - size of a chunk of some slices along an axis  
    
    chunk_shape : tuple
        shape of hyperslab for hdf5 chunking  

    Example  
    
    .. highlight:: python   
    .. code-block:: python  
    
        from ct_segnet.data_io import DataFile  
        # If fname points to existing hdf5 file  
        dfile = DataFile(fname, tiff = False, data_tag = "dataset_name")

        # read a slice  
        img = dfile.read_slice(axis = 1, slice_idx = 100)  

        # read a chunk  of size 2.0 GB starting at slice_start = 0
        vol, s = dfile.read_chunk(axis = 1, slice_start = 0, max_GB = 2.0)  

        # read a chunk between indices [10, 100], [20, 200], [30, 300] along the respective axes  
        vol = dfile.read_data(slice_3D = [slice(10, 100), slice(20, 200), slice(30,300)])

        # or just read all the data
        vol = dfile.read_full()

    """
    
    def __init__(self, fname, data_tag = None, tiff = False,\
                 chunk_shape = None, chunk_size = None, chunked_slice_size = None,\
                 d_shape = None, d_type = None, VERBOSITY = 1):
        self.fname = fname
        self.data_tag = data_tag # for hdf5 only
        self.VERBOSITY = VERBOSITY
        
        self.tiff_mode = tiff
        self.exists = os.path.exists(self.fname)
        self.chunked_slice_size = chunked_slice_size
        self.chunk_size = chunk_size
        self.chunk_shape = chunk_shape

        if (d_type is not None) & (d_shape is not None):
            self.d_type = d_type # overwrite previously read stats            
            self.d_shape = d_shape              
            self.get_slice_sizes()
            self.est_chunking()   
            _message("\n" + "#"*50 + "\n" + "New dataset will be created in %s: "%("tiff folder" if self.tiff_mode else "hdf5 file") + self.fname.split('/')[-1], self.VERBOSITY > 1)        
        elif self.exists:
            self.get_stats()
            self.get_slice_sizes()

        else:
            raise ValueError("Inputs d_shape and d_type are required for new dataset as file does not already exist.\nPath does not exist:\n%s"%fname)
            
        return
    
    def set_verbosity(self, VERBOSITY):
        """
        """
        self.VERBOSITY = val
        return

    def create_new(self, overwrite = False):
        """
        For hdf5 - creates an empty dataset in hdf5 and assigns shape, chunk_shape, etc. For tiff folder - checks if there is existing data in folder.  
        
        :param bool overwrite: if True, remove existing data in the path (fname).
        
        """
        if self.tiff_mode:
            if os.path.exists(self.fname):
                if overwrite:
                    shutil.rmtree(self.fname)
                    _message("Removed old contents in tiff folder %s"%(self.fname.split('/')[-1]), self.VERBOSITY > 0)
                else:
                    raise ValueError("tiff folder already exists and overwrite is not allowed.")
        else:
            if self.data_tag is None:
                raise ValueError("data_tag: argument missing for creating new hdf5 file.")
            if os.path.exists(self.fname):
                with h5py.File(self.fname, 'r') as hf:
                    if self.data_tag in hf.keys():
                        if overwrite:
                            os.remove(self.fname)
                            _message("Removed old hdf5 file %s"%(self.fname.split('/')[-1]), self.VERBOSITY > 0)
                        else:
                            raise ValueError("hdf5 dataset already exists and overwrite is not allowed.")
                    else:
                        pass

            self.est_chunking()
            _opt = 'r+' if os.path.exists(self.fname) else 'w'
            with h5py.File(self.fname, _opt) as hf:
                hf.create_dataset(self.data_tag, shape = self.d_shape, dtype = self.d_type, chunks = self.chunk_shape)
            _message("New hdf5 dataset %s created in file %s"%(self.data_tag, self.fname.split('/')[-1]), self.VERBOSITY > 0)
            
    def get_stats(self, return_output = False):
        """
        Print some stats about the DataFile (shape, slice size, chunking, etc.)  
        
        """
        if not self.tiff_mode:
            with h5py.File(self.fname, 'r') as hf:
                self.d_shape = hf[self.data_tag].shape
                self.d_type = hf[self.data_tag].dtype
                self.chunk_shape = hf[self.data_tag].chunks
        elif self.tiff_mode:
            img_path = glob.glob(self.fname+'/*.tif')
            if not img_path: img_path = glob.glob(self.fname+'/*.tiff')
            img_z = len(img_path)
            img = imread(img_path[0])
            self.d_shape = (img_z,) + img.shape
            self.d_type = img.dtype
            self.chunk_shape = None # no such attribute for tiff stacks
        
        
        str_out = "\n" + "#"*50 + "\n" + "Found existing %s: "%("tiff folder" if self.tiff_mode else "hdf5 file") + self.fname.split('/')[-1]
        str_out = str_out + "\nDataset shape: %s"%(str(self.d_shape))
        
        if return_output:
            return str_out
        else:
            _message(str_out, self.VERBOSITY > 0)        
            
        return

    def get_slice_sizes(self):
        """  
        """
        if self.d_type == np.float32:
            fac = 4.0
        elif self.d_type == np.uint8:
            fac = 1.0
        elif self.d_type == np.uint16:
            fac = 2.0
        else:
            raise ValueError("Data type %s is not supported."%self.d_type)

        self.slice_size = (np.prod(self.d_shape[1:])*fac/(1e9), np.prod(self.d_shape[::2])*fac/(1e9), np.prod(self.d_shape[:-1])*fac/(1e9))
        self._bytes_per_voxel = fac
        self.d_size_GB = fac*np.prod(self.d_shape)/1e9
        return


    def show_stats(self, return_output = False):
        """print dataset shape and slice-wise size
        """
        str_out = ""
        str_out = str_out + "\n" + "Dataset shape: %s"%(str(self.d_shape))
        str_out = str_out + "\n" + "Dataset size: %.2f GB"%self.d_size_GB
        
        if not self.tiff_mode:
            str_out = str_out + "\n" + "Chunk shape: %s"%(str(self.chunk_shape))
            
        for _i, _size in enumerate(self.slice_size):
            str_out = str_out + "\n" + "Slice size along %i: %4.2f MB"%(_i, 1000.0*_size)
        
        if return_output:
            return str_out
        else:
            _message(str_out, self.VERBOSITY > -1)
            return

        
    def est_chunking(self): # Determine the chunk shape for hdf5 file, optimized for slicing along all 3 axes
        """  
        """
        
        if self.tiff_mode:
            self.chunked_shape = None
        else:
            if self.chunk_shape is not None:
                return # allow user to define chunk_shape
            if self.chunk_size is not None:
                fac = np.cbrt((self.chunk_size) / (self.d_size_GB))
                self.chunk_shape = tuple([int(self.d_shape[i]*fac) for i in range(3)])
                return
            if self.chunked_slice_size is not None:
                if self.chunked_slice_size > (self.slice_size[0]*self.d_shape[0]):
                    raise ValueError("chunked_slice_size cannot be larger than dataset size")
                self.chunk_shape = tuple(int(np.ceil(self.chunked_slice_size / single_slice_size)) for single_slice_size in self.slice_size)
                _message("Estimated Chunk shape: %s"%str(self.chunk_shape), self.VERBOSITY > 1)
                return
            else:
                self.chunk_shape = None
                return
        return

    def read_slice(self, axis = None, slice_idx = None):
        """Read a slice.  
        
        :param int axis: axis 0, 1 or 2  
        
        :param int slice_idx: index of slice along given axis  
        
        """
        img, s = self.read_chunk(axis = axis, slice_start = slice_idx, slice_end = slice_idx + 1)
        return img[0]

    def read_data(self, slice_3D = (slice(None,None),)*3):
        """Read a block of data. Only supported for hdf5 datasets.  
        
        :param list slice_3D: list of three python slices e.g. [slice(start,stop,step)]*3
        
        """
        
        if self.tiff_mode:
            ch = np.asarray(read_tiffseq(self.fname, s = slice_3D[0]))
            ch = ch[:, slice_3D[1], slice_3D[2]]
        
        with h5py.File(self.fname, 'r') as hf:
            
            _message("Reading hdf5: %s, Z: %s, Y: %s, X: %s"%(self.fname.split('/')[-1], str(slice_3D[0]), str(slice_3D[1]), str(slice_3D[2])), self.VERBOSITY > 1)    
            ch = np.asarray(hf[self.data_tag][slice_3D[0],slice_3D[1],slice_3D[2]])
        return ch

    def read_sequence(self, idxs):
        """Read a list of indices idxs along axis 0.  
        
        :param int axis: axis 0, 1 or 2  
        
        :param list idxs: list of indices  

        """
        with h5py.File(self.fname, 'r') as hf:
            return np.asarray(hf[self.data_tag][idxs,...])
    
    def read_chunk(self, axis = None, slice_start = None, chunk_shape = None, max_GB = 10.0, slice_end = "", skip_fac = None):
        """Read a chunk of data along a given axis.  
        
        Parameters
        ----------
        
        axis : int
            axis > 0 is not supported for tiff series  
        
        slice_start : int
            start index along axis  
        
        chunk_shape : tuple
            (optional) used if hdf5 has no attribute chunk_shape  
        
        max_GB : float
            maximum size of chunk that's read. slice_end will be calculated from this.  
        
        slice_end : int
            (optional) used if max_GB is not provided.  
        
        skip_fac : int
            (optional) "step" value as in slice(start, stop, step)  
        
        Returns
        -------
        tuple
            (data, slice) where data is a 3D numpy array  
        
        
        """
        if slice_end == "":
            # Do this to determine slice_end based on a max_GB value as RAM limit
            if (chunk_shape is None) & (not self.tiff_mode):
                chunk_shape = self.chunk_shape
            chunk_len = chunk_shape[axis] if chunk_shape is not None else 1
            slice_len = max(1,int(np.round(max_GB/self.slice_size[axis])//chunk_len))*chunk_len
            slice_end =  min(slice_len + slice_start , self.d_shape[axis])
        elif slice_end is None:
            slice_end = self.d_shape[axis]
        
        s = slice(slice_start, slice_end, skip_fac)
        
        if not self.tiff_mode: # hdf5 mode
            with h5py.File(self.fname, 'r') as hf:
                
                _message("Reading hdf5: %s, axis %i,  %s, with chunk_shape: %s"%(self.fname.split('/')[-1], \
                                                                                 axis, s, chunk_shape), self.VERBOSITY > 1)    
                if axis == 0:
                    ch = np.asarray(hf[self.data_tag][s,:,:])
                elif axis == 1:
                    ch = np.asarray(hf[self.data_tag][:,s,:])
                elif axis == 2:
                    ch = np.asarray(hf[self.data_tag][:,:,s])
                ch = np.moveaxis(ch, axis, 0)
            return ch, s
        
        else: # in tiff mode
            if axis != 0: raise ValueError("TIFF data format does not support multi-axial slicing.")

            _message("Reading tiff: %s, axis %i,  %s, chunk_shape: %s"%(self.fname.split('/')[-1],\
                                                                        axis, s, chunk_shape), self.VERBOSITY > 1)    
            ch = np.asarray(read_tiffseq(self.fname, s = s))
            return ch, s

    def read_full(self, skip_fac = None):
        """Read the full dataset
        """
        
        ch, s = self.read_chunk(axis = 0, slice_start = 0, slice_end = self.d_shape[0], skip_fac = skip_fac)
        return ch
    
    def write_full(self, ch):
        """Write the full dataset to filepath.  
        
        :param ch: 3D numpy array to be saved
        
        """
        self.write_chunk(ch, axis = 0, s = slice(0, self.d_shape[0]))
        return        

    def write_data(self, ch, slice_3D = None):
        """Write a block of data. Only supported for hdf5 datasets.  
        
        :param ch: 3D numpy array to be saved  
        
        :param list slice_3D: list of three python slices e.g. [slice(start,stop,step)]*3 - must match shape of ch  
        
        """
        
        if self.tiff_mode:
            raise ValueError("Not supported for tiff data")

        with h5py.File(self.fname, 'r+') as hf:
    
            _message("Saving hdf5: %s, Z: %s, Y: %s, X: %s"%(self.fname.split('/')[-1], str(slice_3D[0]), str(slice_3D[1]), str(slice_3D[2])), self.VERBOSITY > 1)    

            hf[self.data_tag][slice_3D[0],slice_3D[1],slice_3D[2]] = ch
            _message("Done", self.VERBOSITY > 1)
        return
        
        
    def write_chunk(self, ch, axis = None, s = None):
        """Write a chunk of data along a given axis.  
        
        :param int axis: axis > 0 is not supported for tiff series  
        
        :param slice s: python slice(start, stop, step) - step must be None for tiff series  
        
        """
        if not self.tiff_mode:
                
            if not os.path.exists(self.fname):
                raise ValueError("hdf5 file needs to exist before writing chunks.")
            
            with h5py.File(self.fname, 'r+') as hf:
        
                _message("Saving %s, axis %i,  slice %i to %i"%(self.fname.split('/')[-1], axis, s.start, s.stop), self.VERBOSITY > 1)    

                ch = np.moveaxis(ch, 0, axis)
                if axis == 0:
                    hf[self.data_tag][s,:,:] = ch
                elif axis == 1:
                    hf[self.data_tag][:,s,:] = ch
                elif axis == 2:
                    hf[self.data_tag][:,:,s] = ch
                _message("Done", self.VERBOSITY > 1)
    
        else:
            if axis != 0:
                raise ValueError("TIFF data format does not support multi-axial slicing.")
            if os.path.exists(self.fname) & (s.start == 0):
                raise FileExistsError("TIFF folder to be written already exists. Please delete first.")
            _message("Saving %s, axis %i,  slice %i to %i"%(self.fname.split('/')[-1], axis, s.start, s.stop), self.VERBOSITY > 1)    
            
            ch = ch.astype(self.d_type)
            write_tiffseq(ch, SaveDir = self.fname, increment_flag = True,\
                          suffix_len = len(str(self.d_shape[axis])))
        
        return

#### END Class DataFile #######

def read_tiffseq(userfilepath = '', procs = None, s = None):
    """Read a sequence of tiff images from folder.  
    
    :param str userfilepath: path to folder containing images  
    
    :param s: s is either a slice(start, stop, step) or a list of indices to be read  
    
    :type s: slice or list
    
    """
    if not userfilepath:
        raise ValueError("File path is required.")
        return []

    if procs == None:
        procs = cpu_count()
        
    ImgFileList = sorted(glob.glob(userfilepath+'/*.tif'))
    
    if not ImgFileList: ImgFileList = sorted(glob.glob(userfilepath+'/*.tiff'))
    
    if s != None:
        if type(s) == slice:
            ImgFileList = ImgFileList[s]
        elif type(s) == list:
            ImgFileList = [ImgFileList[i] for i in s]
        else:
            raise ValueError("s input not recognized.")
    
    S = Parallelize(ImgFileList, imread, procs = procs)

    return S

def write_tiffseq(S, SaveDir = "", increment_flag = False,\
                  suffix_len = None):
    """Write a sequence of tiff images to a directory.  
    
    :param S: numpy array (3D), sequence will be created along axis 0  
    
    :param str SaveDir: path to folder, will create directory if doesn't exist  
    
    :param bool increment_flag: True to write append images to existing ones in folder  
    
    :param int suffix_len: e.g. 4 for 1000 images, 5 for 10,000  
    
    """
    if not suffix_len:
        if increment_flag:
            raise ValueError("suffix_len required if increment_flag is True.")
        else:
            suffix_len = len(str(S.shape[0]))
    
    last_num = 0
    if not os.path.exists(SaveDir):
        os.makedirs(SaveDir)
    else:
        if not increment_flag:
            shutil.rmtree(SaveDir)
            os.makedirs(SaveDir)
        else:
            ImgFileList = sorted(glob.glob(SaveDir+'/*.tif'))
            if not ImgFileList: ImgFileList = sorted(glob.glob(SaveDir+'/*.tiff'))
            if not ImgFileList:
                last_num = 0
            else:
                last_num = int(ImgFileList[-1].split('.')[0][-suffix_len:])    
    
    BaseDirName = os.path.basename(os.path.normpath(SaveDir))
    
    for iS in range(S.shape[0]):
        img_num = str(iS+1+last_num).zfill(suffix_len)
        imsave(os.path.join(SaveDir, BaseDirName + img_num + '.tif'), S[iS])
    
    return
    
def Parallelize(ListIn, f, procs = -1, **kwargs):
    
    """This function packages the "starmap" function in multiprocessing, to allow multiple iterable inputs for the parallelized function.  
    
    :param list ListIn: list, each item in the list is a tuple of non-keyworded arguments for f.  
    
    :param func f: function to be parallelized. Signature must not contain any other non-keyworded arguments other than those passed as iterables.  
    
    Example:
    
    .. highlight:: python
    .. code-block:: python
    
        def multiply(x, y, factor = 1.0):
            return factor*x*y
    
        X = np.linspace(0,1,1000)  
        Y = np.linspace(1,2,1000)  
        XY = [ (x, Y[i]) for i, x in enumerate(X)] # List of tuples  
        Z = Parallelize_MultiIn(XY, multiply, factor = 3.0, procs = 8)  
    
    Create as many positional arguments as required, but remember all must be packed into a list of tuples.
    
    """
    if type(ListIn[0]) != tuple:
        ListIn = [(ListIn[i],) for i in range(len(ListIn))]
    
    reduced_argfunc = functools.partial(f, **kwargs)
    
    if procs == -1:
        opt_procs = int(np.interp(len(ListIn), [1,100,500,1000,3000,5000,10000] ,[1,2,4,8,12,36,48]))
        procs = min(opt_procs, cpu_count())

    if procs == 1:
        OutList = [reduced_argfunc(*ListIn[iS]) for iS in range(len(ListIn))]
    else:
        p = Pool(processes = procs)
        OutList = p.starmap(reduced_argfunc, ListIn)
        p.close()
        p.join()
    
    return OutList
   
def show_header():
    print("\n" + "#"*60 + "\n")
    print("\tWelcome to CTSegNet: AI-based 3D Segmentation")
    print("\n" + "#"*60 + "\n")
    return

def show_endmessage(_str):
    print("\n" + "#"*60 + "\n")
    print("\t" + _str)
    print("\n" + "#"*60 + "\n")
    return
    

def _istiff(fpath, data_tag):
    """
    Returns True if path provided is to a directory of tiffs, False if .hdf5 file.
    Raises ArgumentTypeError if format is not supported.
    """
    # Understand input data format
    if os.path.isdir(fpath):
        tiff_input = True
    elif fpath.split('.')[-1] in ("hdf5", "h5"):
        tiff_input = False
        if data_tag == "":
            raise ArgumentTypeError("dataset-name required for hdf5")
    else:
        raise ArgumentTypeError("input file type not recognized. must be tiff folder or hdf5 file")
    return tiff_input

def get_cropped_shape(crops, d_shape):
    if crops is not None:
        crop_shape = [0]*len(d_shape)
        for idx, crop in enumerate(crops):
            _size = [0,0]
            if (crop[0] is not None):
                if crop[0] >= 0:
                    _size[0] = min(abs(crop[0]), d_shape[idx])
                elif crop[0] < 0:
                    _size[0] = max(0, d_shape[idx] - abs(crop[0]))
            else:
                _size[0] = 0

            if crop[1] is not None:
                if crop[1] >= 0:
                    _size[1] = min(abs(crop[1]), d_shape[idx])
                elif crop[1] < 0:
                    _size[1] = max(0, d_shape[idx] - abs(crop[1]))
            else:
                _size[1] = d_shape[idx]
            crop_shape[idx] = max(_size[1] - _size[0], 0)
    else:
        crop_shape = d_shape
    return tuple(crop_shape)
        
    
    
if __name__ == "__main__":
    
    mem_thres = 20.0
    _message("Data input / output functions.")






#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Post-process tomo data after passing through CTSegNet. Do this to remove tessellation artifacts.

This file can be run as a standalone executable or the "morpho_filter" function can be imported into your code.

"""
import time
import sys
import numpy as np


from ct_segnet.data_utils.data_io import DataFile, crops_type, str_to_bool
from ct_segnet.seg_utils import get_repadding

from scipy.ndimage import binary_dilation, binary_erosion
import configargparse


def make_binary_structure(radius, ndim = 3):
    struct_size = 2*radius + 1
    struct = np.ones(tuple([struct_size]*ndim))

    r = np.arange(-radius, radius + 1)

    yy, xx = np.meshgrid(r, r, indexing = 'ij')
    dist = np.sqrt(yy**2 + xx**2)
    struct[dist > radius] = 0
    return struct


def morpho_filter(vol, ops = ["erode", "dilate"], radius = [5, 5], output = None, crops = None, invert_mask = False):
    """
    this removes tessellation artifacts
    Parameters
    ----------
    vol : np.array
        binary (labeled) volume; can also be a 2D image
    ops : list
        list of operations - pick from "erode" and "dilate"
    radius : radius or list of radii for the respective operations
    invert_mask : bool
        boolean True if mask must be inverted before filters applied
    
    Returns
    -------
    np.array
        filtered binary (labeled) volume
    """

    if invert_mask:
        _message("Inverting mask...")
        vol = vol^1 # invert the labels if required by user. Some models interpret features as voids rather than particles and vice versa.
    
    if crops is not None:
        pads = get_repadding(crops, vol.shape)
        vol = vol[slice(*crops[0]), slice(*crops[1]), slice(*crops[2])]
    
    print("Morphological operations: ")
    if type(radius) is not list:
        radius = [radius]*len(ops)
    
    t00 = time.time()
    for idx, op in enumerate(ops):
        t0 = time.time()
        struct = make_binary_structure(radius[idx], ndim = vol.ndim)
        
        print("%s (r = %i),"%(op, radius[idx]), end = " ")
        if op == "erode":
            vol = binary_erosion(vol, structure = struct, output = output)
        elif op == "dilate":
            vol = binary_dilation(vol, structure = struct, output = output)
        t_step = time.time() - t0
        print("%.1f secs; "%t_step, end = " ")
    
    # Pad the volume to bring it back to original dimensions
    if crops is not None:
        vol = np.pad(vol, pads, 'constant', constant_values = 0)
    
    tot_time = (time.time() - t00)/60.0
    print("\ntotal time for morphological ops = %.2f minutes"%tot_time)
    return vol



def _message(_str):
    print(_str)
    return
    
    
def main(args):
    
    # Create an instance of DataFile of the existing recon
    ds_seg = DataFile(args.input_fname, tiff = False, data_tag = "SEG", VERBOSITY = 0)
    
    
    # Save the output mask to a new file
    ds2 = DataFile(args.output_fname, \
                   data_tag = "SEG", \
                   tiff = False, \
                   d_shape = ds_seg.d_shape, \
                   d_type = ds_seg.d_type, \
                   chunk_shape = ds_seg.chunk_shape,\
                   VERBOSITY = 0)
    ds2.create_new(overwrite = True)
    
    t00 = time.time()
    print("Reading dataset: %s"%ds_seg.fname)
    vol = ds_seg.read_full()
    
    vol = morpho_filter(vol, radius = args.radius, \
                        ops = args.ops, \
                        crop = args.crops, \
                        invert_mask = args.invert_mask)
    
    
    # Save the volume back to the designated DataFile object
    ds2.write_full(vol)
    tot_time_rw = (time.time() - t00)/60.0
    
    print("Done. Total elapsed time (include r/w): %.2f minutes"%tot_time_rw)


if __name__ == "__main__":
    
    
    parser = configargparse.ArgParser()
    
    # list of arguments accepted
    parser.add('-f', '--input_fname', required = True, type = str, help = 'input file name (.hdf5)')
    parser.add('-o', '--output_fname', required = True, type = str, help = 'output file name (.hdf5)')
    parser.add('--ops', required = True, type = str, action = 'append')
    parser.add('--radius', required = True, type = int, action = 'append')
    parser.add('--crops', type = crops_type, action = 'append')
    parser.add('--invert_mask', required = False, type = str_to_bool, default = True, metavar = 'bool', help = 'True to invert mask before applying ops')
    
               
    args = parser.parse_args()
    main(args)
    


    
    
    
    
    
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Example functions for measuring stuff from segmentation masks. Pass these to the FeatureExtraction2D class.  


"""
import numpy as np
import os
import pandas as pd
from ct_segnet import stats


def segmentation_accuracy(seg_img, true_img = None, metrics = ["jaccard", "dice"]):
    
    '''  
    Calculates segmentation accuracy metrics when provided with a ground-truth map.  
    Returns
    -------
    
    np.array  
        vector of accuracy metrics
    
    Parameters
    ----------
    
    seg_img : np.array  
        segmented image (output from Segmenter class)  
        
    true_img : np.array  
        true segmentation map (same shape as seg_img)  
        
    metrics : list  
        list of metric names from "jaccard", "dice", "dice_edge"  
    
    '''
    
    values = []
    for metric in metrics:
        
        if metric == "dice":
            values.append(stats.calc_dice_coeff(true_img, seg_img))
        elif metric == "jaccard":
            values.append(stats.calc_jac_acc(true_img, seg_img))
        
    return np.asarray(values)

from scipy.ndimage import label, find_objects

def pore_analysis(seg_img, features = ["fraction", "number", "size"], invert = False):
    
    '''
    Calculates void fraction, number of pores and mean size of pores. Pore size is simply defined by sqrt of sum of pixels in the pore phase.  
    
    Returns  
    -------  
    np.array  
        vector of measurements  
    
    Parameters   
    ----------  
    seg_img : np.array  
        segmented image (output from Segmenter class)  
        
    true_img : np.array  
        true segmentation map (same shape as seg_img)  
        
    features : list  
        list of porosity features to measure from "fraction", "number" and "size"    
        
    invert : bool  
        if the image contains particles instead of pores, invert = True  
        
    '''
    
    if any(f in features for f in ["number", "size"]):
        img_labeled, n_objs = label(seg_img if invert else seg_img^1)
        obj_list = find_objects(img_labeled)
    
    values = []
    for feature in features:
        if feature == "fraction":
            values.append(1 - np.sum(seg_img)/np.size(seg_img))
        elif feature == "number":
            values.append(n_objs)
        elif feature == "size":
            
            p_size = []
            for idx in range(n_objs):

                sub_img = img_labeled[obj_list[idx]]
                p_area = np.sum(sub_img==(idx+1))                
                p_dia_px = np.sqrt(p_area)#*2*3/(4*np.pi)
                p_size.append(p_dia_px)

            values.append(np.mean(p_size))
        
    return np.asarray(values)

    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
CTSegNet is more than a 2D CNN model - it's a 3D Segmenter that uses 2D CNNs. The set_utils.py defines the Segmenter class that wraps over a keras U-net-like model (defined by models.py), integrating 3D slicing and 2D patching functions to enable the 3D-2D-3D conversations in the segmentation workflow. To slice a 3D volume, manipulations such as 45 deg rotations, orthogonal slicing, patch extraction and stitching are performed.

"""

import sys
import os
# line 13 empty for good luck
import numpy as np
import pandas as pd
import re
import ast
import h5py
import cv2
import time
import tensorflow as tf
from tensorflow.keras.models import load_model

from ct_segnet.data_utils import patch_maker as PM
from ct_segnet.data_utils.data_io import Parallelize
from ct_segnet.model_utils.losses import custom_objects_dict



VERBOSE = False

def message(_str):
    
    if VERBOSE:
        print(_str)
    return


class Segmenter():
    """
    The Segmenter class wraps over a keras model, integrating 3D slicing and 2D patching functions to enable the 3D-2D-3D conversations in the segmentation workflow.  
    
    model: tf.keras.model  
        keras model with input shape = out shape = (ny, nx, 1)  
    
    model_filename : str  
        path to keras model file (e.g. "model_1.h5")  
    
    model_name : str  
        (optional) just a name for the model  
    
    """
    def __init__(self, model_filename = None, model = None, model_name = "unknown"):
        if model is not None:
            self.model = model
            self.model_name = model_name
        else:
            self.model_name = os.path.split(model_filename)[-1].split('.')[0]
            self.model = load_model(model_filename, custom_objects = custom_objects_dict)

    def seg_image(self, s, max_patches = None, overlap = None):

        """
        Test the segmenter on arbitrary sized 2D image. This method extracts patches of shape same as the input shape of 2D CNN, segments them and stitches them back to form the original image.  
        
        max_patches : tuple  
            (my, mx) are # of patches along Y, X in image  
        
        s : numpy.array  
            greyscale image slice of shape (ny, nx)  
        
        overlap : tuple or int  
            number of overlapping pixels between patches  
        
        """
        # Handle patching parameter inputs
        patch_size = self.model.output_shape[1:-1]    
        if type(max_patches) is not tuple:
            max_patches = (max_patches, max_patches)
            
        if type(overlap) is not tuple:
            overlap = (overlap, overlap)
        overlap = (0 if max_patches[0] == 1 else overlap[0],\
                   0 if max_patches[1] == 1 else overlap[1])
    
        # Resize images
        orig_shape = s.shape
        s = cv2.resize(s, (max_patches[1]*patch_size[1] - overlap[1],\
                           max_patches[0]*patch_size[0] - overlap[0]))
        
        # Make patches
        downres_shape = s.shape
        steps = PM.get_stepsize(downres_shape, patch_size)
        s = PM.get_patches(s, patch_size = patch_size, steps = steps)
        
        # The dataset now has shape: (ny, nx, py, px). ny, nx are # of patches, and py, px is patch_shape.
        # Reshape this dataset into (n, py, px) where n = ny*nx. Trust numpy to preserve order. lol.
        dataset_shape = s.shape
        s = s.reshape((-1,) + patch_size)
        
        # Predict using the model.
        s = self.model.predict(s[...,np.newaxis])
        s = s[...,0]
        
        # Now, reshape the data back...
        s = s.reshape(dataset_shape)
        
        # Reconstruct from patches...
        s = PM.recon_patches(s, img_shape = downres_shape, steps = steps)
        
        # Finally, resize the images to the original shape of slices... This will result in some loss of resolution...
        s = cv2.resize(s, (orig_shape[1], orig_shape[0]))

        # outputs: segmented image of same shape as input image p
        return np.asarray(np.round(s)).astype(np.uint8)  
    

    def seg_chunk(self, p, max_patches = None, overlap = None,\
                  nprocs = None, arr_split = 1, arr_split_infer = 1):
        """
        Segment a volume of shape (nslices, ny, nx). The 2D keras model passes\
        along nslices, segmenting images (ny, nx) with a patch size defined by input \
        to the model  
        
        max_patches: tuple  
            (my, mx) are # of patches along Y, X in image (ny, nx)  

        overlap : tuple or int  
            number of overlapping pixels between patches  
        
        nprocs : int  
            number of CPU processors for multiprocessing Pool  
        
        arr_split : int  
            breakdown chunk into arr_split number of smaller chunks  
        
        """
        
        # Handle patching parameter inputs
        patch_size = self.model.output_shape[1:-1]
        if type(max_patches) is not tuple:
            max_patches = (max_patches, max_patches)
            
        if type(overlap) is not tuple:
            overlap = (overlap, overlap)
        overlap = (0 if max_patches[0] == 1 else overlap[0],\
                   0 if max_patches[1] == 1 else overlap[1])
    
        # Resize images
        orig_shape = p[0].shape
        p = np.asarray([cv2.resize(p[ii], (max_patches[1]*patch_size[1] - overlap[1],\
                                           max_patches[0]*patch_size[0] - overlap[0]))\
                        for ii in range(p.shape[0])])
        
        # Make patches
        message("Making patches...")
        message("\tCurrent d shape:" + str(np.shape(p)))
        downres_shape = p[0].shape
        steps = PM.get_stepsize(downres_shape, patch_size)
        p = Parallelize(p, PM.get_patches, procs = nprocs, \
                           patch_size = patch_size, steps = steps)
        p = np.asarray(p)
        
        # The dataset now has shape: (nslices, ny, nx, py, px),
        # where ny, nx are # of patches, and py, px is patch_shape.
        # Reshape this dataset into (n, py, px) where n = nslices*ny*nx.
        dataset_shape = p.shape
        p = p.reshape((-1,) + patch_size)
        
        # Predict using the model.
        message("Running predictions using model...")
        message("\tCurrent d shape:" + str(np.shape(p)))
        
        
        p = np.array_split(p, arr_split_infer)
        for jj in range(len(p)):
            p[jj] = self.model.predict(p[jj][...,np.newaxis])[...,0]
            p[jj] = np.round(p[jj])
        p = np.concatenate(p, axis = 0)
        
        p = p.astype(np.uint8) # typecasting
        
        # Now, reshape the data back...
        p = p.reshape(dataset_shape)
        p = [p[ii] for ii in range(p.shape[0])]
        
        # Reconstruct from patches...
        message("Reconstructing from patches...")
        message("\tCurrent d shape:" + str(np.shape(p)))
        p = np.array_split(p, arr_split)
        
        p = [np.asarray(Parallelize(p[ii], PM.recon_patches,\
                                    img_shape = downres_shape,\
                                    steps = steps, procs = nprocs\
                                   )) for ii in range(arr_split)]
        p = np.concatenate(p, axis = 0)
        
        # Finally, resize the images to the original shape of slices... This will result in some loss of resolution...
        message("Resizing images to original slice size...")
        message("\tCurrent d shape:" + str(np.shape(p)))
        p = np.asarray([cv2.resize(p[ii], (orig_shape[1], orig_shape[0]))\
                        for ii in range(p.shape[0])])
        return p

def get_repadding(crops, d_shape):

    """
    
    Returns  
    -------  
        tuple
            padding values to restore 3D np array after it was cropped.  
    
    Parameters  
    ----------  
    
    crops : list
        3 tuples in a list [(nz1,nz2), (ny1,ny2), (nx1,nx2)]  

    d_shape : tuple
        original shape of 3D array  
    
    """
    pads = []
    for idx, crop in enumerate(crops):
        pad = [0,0]
        if (crop[0] is not None):
            if crop[0] >= 0:
                pad[0] = abs(crop[0])
            elif crop[0] < 0:
                pad[0] = d_shape[idx] - abs(crop[0])
        
        if crop[1] is not None:
            if crop[1] >= 0:
                pad[1] = d_shape[idx] - abs(crop[1])
            elif crop[1] < 0:
                pad[1] = abs(crop[1])
        pads.append(tuple(pad))
        
    return tuple(pads)

                
def _rotate(imgs, angle):
    """  
    
    Just a wrapper for cv2's affine transform for rotating an image about center  
    
    Parameters  
    ----------  
    
    imgs : np.array  
        volume or series of images (n, ny, nx)  
    
    angle : float  
        value to rotate image about center, along (ny,nx)  
    
    """
    rows, cols = imgs[0].shape
    M = cv2.getRotationMatrix2D((cols/2,rows/2), angle,1)
    return np.asarray([cv2.warpAffine(imgs[iS],M,(cols,rows)) for iS in range(len(imgs))])    

def process_data(p, segmenter, preprocess_func = None, max_patches = None,\
                 overlap = None, nprocs = None, rot_angle = 0.0, slice_axis = 0,\
                 crops = None, arr_split = 1, arr_split_infer = 1):
    """
    Segment a volume of shape (nz, ny, nx). The 2D keras model passes
    along either axis (0,1,2), segmenting images with a patch size defined by input
    to the model in the segmenter class.  

    Parameters
    ----------
    max_patches : tuple  
        (my, mx) are # of patches along Y, X in image (ny, nx)  

    overlap : tuple or int  
        number of overlapping pixels between patches  

    nprocs : int  
        number of CPU processors for multiprocessing Pool  
    
    arr_split : int  
        breakdown chunk into arr_split number of smaller chunks  
        
    slice_axis : int  
        (0,1,2); axis along which to draw slices  
    

    crops : list  
        list of three tuples; each tuple (start, stop) will define a python slice for the respective axis  
    
    rot_angle : float  
        (degrees) rotate volume around Z axis before slicing along any given axis. Note this is redundant if slice_axis = 0  
        
                      
    nprocs : int  
        number of CPU processors for multiprocessing Pool  
   
    arr_split : int  
        breakdown chunk into arr_split number of smaller chunks  
    
    preprocess_fun : func  
        pass a preprocessing function that applies a 2D filter on an image  
        
    """
    if nprocs is None:
        nprocs = 4
    if p.ndim != 3:
        raise ValueError("Invalid dimensions for 3D data.")

    message("Orienting, rotating and padding as requested...")
    # Rotate the volume along axis 0, if requested
    if rot_angle > 0.0:
        p = _rotate(p, rot_angle)
    
    if crops is not None:
        pads = get_repadding(crops, p.shape)
        p = p[slice(*crops[0]), slice(*crops[1]), slice(*crops[2])]

    # Orient the volume such that the first axis is the direction in which to slice through...
    p = np.moveaxis(p, slice_axis, 0)
    message("\tDone")

    # Preprocess function
    if preprocess_func is not None:
#         print("\tPreprocessing on XY mapping...")
        p = preprocess_func(p)

    # Run the segmenter algorithm
    p = segmenter.seg_chunk(p, max_patches = max_patches, overlap = overlap, nprocs = nprocs, arr_split = arr_split, arr_split_infer = arr_split_infer)


    message("Re-orienting, rotating and padding back original size...")
    # Re-orient the volume such that the first axis is the vertical axis...
    p = np.moveaxis(p, 0, slice_axis)
    

    # Pad the volume to bring it back to original dimensions
    if crops is not None:
        p = np.pad(p, pads, 'constant', constant_values = 0)
    
    # Rotate the volume along axis 0, back to its original state
    if rot_angle > 0.0:
        p = _rotate(p, -rot_angle)
    message("\tDone")
    
    return p.astype(np.uint8)



class FeatureExtraction2D(Segmenter):
    '''  
    This class converts a 2D image into an n-dimensional vector z  
    
    Parameters  
    ----------  
    
    max_patches : tuple  
        (my, mx) are # of patches along Y, X in image  


    overlap : tuple or int  
        number of overlapping pixels between patches  
    
    model: tf.keras.model  
        keras model with input shape = out shape = (ny, nx, 1)  
    
    model_filename : str  
        path to keras model file (e.g. "model_1.h5")  
    
    model_name : str  
        (optional) just a name for the model  
    
    '''  
    
    def __init__(self, max_patches = None, overlap = None, model_filename = None):

        
        if type(max_patches) is not tuple:
            max_patches = (max_patches, max_patches)
            
        if type(overlap) is not tuple:
            overlap = (overlap, overlap)
        overlap = (0 if max_patches[0] == 1 else overlap[0],\
                   0 if max_patches[1] == 1 else overlap[1])
        
        self.max_patches = max_patches
        self.overlap = overlap
        self.model_filename = model_filename
#         super(Segmenter, self).__init__(self, model_filename = model_filename)        

        self.model_name = os.path.split(model_filename)[-1].split('.')[0]
        self.model = load_model(model_filename, custom_objects = custom_objects_dict)



    def extract_measurement(self, img, measurement, **kwargs):
        
        '''  
        Returns  
        -------  
        measured_features : np.array  
            shape (ndims,)
        
        Parameters  
        ----------  
        
        img : np.array
            A 2D numpy array (ny,nx). Could be a tomo slice or projection.  
            
        measurement : func  
            function to extract a measurement, e.g. radius, particle centroid, etc.  
            
        '''
        
        if measurement is None:
            raise "ValueError: missing argument measurement is required"
        
        seg_img = self.seg_image(img, max_patches = self.max_patches, overlap = self.overlap)
        
        measured_features = measurement(seg_img, **kwargs)
        return measured_features
        
    def extract_code(s):
        '''
        not implemented  
        
        to do:
        consider patches are created. How should the code vectors of each patch be converted to singe vector? (mean, median, std?)  
        '''  
        raise NotImplementedError
        


    def vis_feature(self, s, measurement, **kwargs):

        """
        This method extracts patches of shape same as the input shape of 2D CNN, measures a feature for each patch's segmentation map and stitches them back to form a checkered image. 
        
        s : numpy.array  
            greyscale image slice of shape (ny, nx)  
        
        """
        
        # Handle patching parameter inputs
        patch_size = self.model.output_shape[1:-1]    
    
        # Resize images
        orig_shape = s.shape
        s = cv2.resize(s, (self.max_patches[1]*patch_size[1] - self.overlap[1],\
                           self.max_patches[0]*patch_size[0] - self.overlap[0]))
        
        # Make patches
        downres_shape = s.shape
        steps = PM.get_stepsize(downres_shape, patch_size)
        s = PM.get_patches(s, patch_size = patch_size, steps = steps)
        
        # The dataset now has shape: (ny, nx, py, px). ny, nx are # of patches, and py, px is patch_shape.
        # Reshape this dataset into (n, py, px) where n = ny*nx. Trust numpy to preserve order.
        dataset_shape = s.shape
        s = s.reshape((-1,) + patch_size)
        
        # Predict using the model.
        s = self.model.predict(s[...,np.newaxis])
        s = s[...,0]
        
        s = np.round(s).astype(np.uint8)
        
        f = [measurement(s[idx], **kwargs) for idx in range(len(s))]
        f = np.asarray(f)
        # the shape of f now is n_imgs, n_features
        nimgs, nfeatures = f.shape
        
        s = np.ones((nimgs, patch_size[0], patch_size[1], nfeatures))
        
        for ife in range(nfeatures):
            s[...,ife] = [s[ii,...,ife]*f[ii,ife] for ii in range(len(s))]
        s = np.asarray(s)
        
        
        
        F_img = []
        for ife in range(nfeatures):
            # Now, reshape the data back...
            f_img = s[...,ife].reshape(dataset_shape)
            # Reconstruct from patches...
            f_img = PM.recon_patches(f_img, img_shape = downres_shape, steps = steps)    
            # Finally, resize the images to the original shape of slices
            f_img = cv2.resize(f_img, (orig_shape[1], orig_shape[0]))            
            F_img.append(f_img)
        
        
        return np.asarray(F_img)
        







if __name__ == "__main__":
    
    message("\n" + "#"*50 + "\n")
    message("Welcome to CTSegNet: AI-based 3D Segmentation\n")
    message("\n" + "#"*50 + "\n")




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
view segmentation maps overlain with image data
"""

# line 13 empty for good luck

import sys
import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from ct_segnet.data_utils import patch_maker as PM
# from ImageStackPy import ImageProcessing as IP

from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar
import matplotlib.font_manager as fm
from skimage.filters import sobel

def view_midplanes(vol = None, ds = None, ax = None, cmap = 'gray', alpha = None, idxs = None, axis_off = False, label_planes = True):
    """View 3 images drawn from planes through axis 0, 1, and 2 at indices listed (idx). Do this for a DataFile or numpy.array
    
    :param matplotlib.axes ax:   three axes  
    
    :param numpy.array vol: 3D numpy array  
    
    :param DataFile ds: If vol is provided, ignore DataFile, else read from this DataFile  
    
    """

    if ax is None:
        fig, ax = plt.subplots(1,3)
    imgs = get_orthoplanes(ds = ds, vol = vol, idxs = idxs)
    for i in range(3):
        ax[i].imshow(imgs[i], cmap = cmap, alpha = alpha)
    
    if label_planes:
        h = ax[0].set_title("XY mid-plane")
        h = ax[1].set_title("XZ mid-plane")
        h = ax[2].set_title("YZ mid-plane")    
    
    if axis_off:
        for ii in range(3):
            ax[ii].axis('off')
    
    return ax

def get_orthoplanes(ds = None, vol = None, idxs = None):
    """Return 3 images drawn from planes through axis 0, 1, and 2 at indices listed (idx). Do this for a DataFile or numpy.array
    
    :return: images at three midplanes  
    
    :rtype: list
    
    :param matplotlib.axes ax:   three axes  
    
    :param numpy.array vol   : 3D numpy array  
    
    :param DataFile ds: If vol is provided, ignore DataFile, else read from this DataFile  
    
    """
    
    if vol is not None:
        if idxs is None: idxs = [vol.shape[i]//2 for i in range(3)]
        imgs = [vol.take(idxs[i], axis = i) for i in range(3)]
    elif ds is not None:
        if idxs is None: idxs = [ds.d_shape[i]//2 for i in range(3)]
        imgs = [ds.read_slice(axis = i, slice_idx = idxs[i]) for i in range(3)]
    
    return imgs    

def edge_plot(img, seg_img, ax, color = [0,255,0]):
    """
    Show edge-map of segmented map overlain on greyscale image.  
    
    :param numpy.array img: grayscale image of any shape (Y,X)  
    
    :param numpy.array seg_img: corresponding segmented image  
    
    :param matplotlib.axes ax: One axis object to plot into  
    
    :param list color: list of the 3 RGB values
    
    """
    
    img = (255*((img - img.min()) / (img.max() - img.min()))).astype(np.uint8)
    img = np.concatenate([img[...,np.newaxis]]*3, axis = -1)
#     sob_img = IP.calc_sobel(np.copy(seg_img))[0]
    sob_img = sobel(np.copy(seg_img))
    sob_img = (sob_img - sob_img.min()) / (sob_img.max() - sob_img.min())
    sob_img = (255*np.round(sob_img)).astype(np.uint8)
    img[sob_img == 255] = color

    ax.imshow(img)
    
    return ax

def seg_plot(img, seg_img, ax, alpha = 0.3, cmap = 'gray'):

    """
    Show edge-map of segmented map overlain on greyscale image.  
    
    :param numpy.array img: grayscale image of any shape (Y,X)  
    
    :param numpy.array seg_img: corresponding segmented image  
    
    :param matplotlib.axes ax: One axis object to plot into  
    
    :param float alpha : blending value
    
    """
    
    
    img = (255*((img - img.min()) / (img.max() - img.min()))).astype(np.uint8)
    ax.imshow(img, cmap = cmap)
    ax.imshow(seg_img, cmap = 'copper', alpha = alpha)
    
    return ax


def add_scalebar(ax, bar_len, resolution, units = 'um', n_dec = 0, pad = 0.35, \
                 fontsize = 16, frameon = True, color = 'black', loc = 'lower center'):
    """
    :param float bar_len: length of bar, e.g. 100 micron  
    
    :param float resolution: pixel size, in units/pixel, e.g. 1.1 microns / pixel  
    
    :param str units: units, e.g. 'um'  
    
    :param int n_dec: number of decimals in number displayed below scalebar  
    
    """

    fontprops = fm.FontProperties(size=fontsize)
    n_pix = bar_len/resolution
    scalebar = AnchoredSizeBar(ax.transData, n_pix, ("%."+ "%if"%n_dec)%bar_len + " " + units, loc, \
                               pad=pad, color=color, frameon=frameon, size_vertical=1, fontproperties=fontprops)
    ax.add_artist(scalebar)
    return ax



if __name__ == "__main__":
    
    print("\n" + "#"*50 + "\n")
    print("Welcome to CTSegNet: AI-based 3D Segmentation\n")
    print("\n" + "#"*50 + "\n")




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Easily define U-net-like architectures using Keras layers

"""

import numpy as np
from tensorflow import keras
from tensorflow.keras import layers as L
from . import losses

def insert_activation(tensor_in, activation):
    """
    :return: tensor of rank 4 (batch_size, n_rows, n_cols, n_channels)
    
    Parameters
    ----------
    tensor_in : tensor
            input tensor
    activation : str or tf.Keras.layers.Activation
            name of custom activation or Keras activation layer
            
    """
    if activation == 'lrelu':
        tensor_out = L.LeakyReLU(alpha = 0.2)(tensor_in)
    else:
        tensor_out = L.Activation(activation)(tensor_in)
    
    return tensor_out
    


def conv_layer(tensor_in, n_filters, kern_size = None, activation = None, kern_init = 'he_normal', padding = 'same', dropout = 0.1, batch_norm = False):

    """
    Define a block of two convolutional layers
    
    :return: tensor of rank 4 (batch_size, n_rows, n_cols, n_channels)
    
    Parameters
    ----------
    tensor_in  : tensor
            input tensor
    n_filters  : int
            number of filters in each convolutional layer
    kern_size  : tuple
            kernel size, e.g. (3,3)
    activation : str or tf.Keras.layers.Activation
            name of custom activation or Keras activation layer
    kern_init  : str
            kernel initialization method
    padding    : str
            type of padding
    dropout    : float
            dropout fraction
    batch_norm : bool
            True to insert a BN layer
            
    """
    
    # layer # 1
    tensor_out = L.Conv2D(n_filters, kern_size, activation = None, kernel_initializer = kern_init, padding = padding)(tensor_in)
    
    if batch_norm:
        tensor_out = L.BatchNormalization(momentum = 0.9, epsilon = 1e-5)(tensor_out)
    
    tensor_out = insert_activation(tensor_out, activation)
    
    # Dropout
    if dropout is not None:
        tensor_out = L.Dropout(dropout)(tensor_out)
    
    # Layer # 2
    tensor_out = L.Conv2D(n_filters, kern_size, activation = None, kernel_initializer = kern_init, padding = padding)(tensor_out)

    if batch_norm:
            tensor_out = L.BatchNormalization(momentum = 0.9, epsilon = 1e-5)(tensor_out)
        
    tensor_out = insert_activation(tensor_out, activation)

    return tensor_out
    
def upconv_layer(tensor_in, concat_tensor, n_filters = None, activation = None, kern_size = None, strides = None, padding = 'same', batch_norm = False):
    """
    Define an upconvolutional layer and concatenate the output with a conv layer from the contracting path
    
    :return: tensor of rank 4 (batch_size, n_rows, n_cols, n_channels)
    
    Parameters
    ----------
    tensor_in     : tensor
            input tensor
    concat_tensor : tensor
            this will be concatenated to the output of the upconvolutional layer
    n_filters  : int
            number of filters in each convolutional layer
    kern_size  : tuple
            kernel size for upconv, e.g. (2,2)
    activation : str or tf.Keras.layers.Activation
            name of custom activation or Keras activation layer
    kern_init  : str
            kernel initialization method
    strides    : tuple
            strides e.g. (2,2)
    padding    : str
            type of padding
    batch_norm : bool
            True to insert a BN layer
    
    """
    if not n_filters:
        n_filters = int(tensor_in.shape[-1]) // 2
    
    tensor_out = L.Conv2DTranspose(n_filters, kern_size, strides = strides, padding = padding, activation = None) (tensor_in)
    tensor_out = L.concatenate([tensor_out, concat_tensor])
    
#    if batch_norm:
#        tensor_out = L.BatchNormalization(momentum = 0.9, epsilon = 1e-5)(tensor_out)
#    
#    tensor_out = insert_activation(tensor_out, activation)
#    
    return tensor_out
    
def pool_layer(tensor_in, n_filters, pool_size, dropout = None, activation = None, batch_norm = False, kern_size = None):
    """
    Define a block of 2 convolutional layer followed by a pooling layer
    
    :return: tensor of rank 4 (batch_size, n_rows, n_cols, n_channels)
    
    Parameters
    ----------
    tensor_in     : tensor
            input tensor
    n_filters  : int
            number of filters in each convolutional layer
    pool_size  : tuple
            max pooling (2,2)
    dropout    : float
            fraction of dropout
    activation : str or tf.Keras.layers.Activation
            name of custom activation or Keras activation layer
    batch_norm : bool
            True to insert a BN layer
    kern_size  : tuple
            kernel size for conv layer, e.g. (3,3)
    
    """
    conv = conv_layer(tensor_in, n_filters, dropout = 0.1, activation = activation, batch_norm = batch_norm, kern_size = kern_size)
    pool = L.MaxPooling2D(pool_size)(conv)
    
    return conv, pool
    
def _expand_inputs(inp, n, var_name):
    """
    Handle inputs provided.  
    
    For n_depth and dropouts, if input is a value, return list with length equal to number of pooling layers.  
    
    For kernel sizes, if input is tuple, make a list of the same tuple, for all pooling layers. If it is a list, ensure the length is same as number of pooling layers.  
    
    """
    # The last n_filter is for bottleneck layer.You can extend this list for U-net with more than 4 pooling layers.
    
    dict_vars = {'n_depth' : [16, 32, 64, 128, 256],\
                 'dropouts': [0.1, 0.1, 0.2, 0.2, 0.3]}
    if var_name in dict_vars.keys():
        if type(inp) is not list:
            inp = np.asarray(dict_vars[var_name])*inp
            inp = inp[:n]
        else:
            if len(inp) != n:
                raise ValueError("Incorrect length for list: %s"%var_name)
    else:
        if type(inp) is tuple:
            inp = [inp]*n
        elif type(inp) is list:
            if len(inp) != n:
                raise ValueError("Incorrect number of elements specified in %s"%var_name)

    return inp
            
    

def build_Unet_flex(img_shape, n_depth = 1, n_pools = 4, activation = 'lrelu', \
                     batch_norm = True, kern_size = (3,3), kern_size_upconv = (2,2), pool_size = (2,2), dropout_level = 1.0,\
                     loss = 'binary_crossentropy', stdinput = True):
    """
    Define your own Unet-like architecture, based on the arguments provided. Checks that the architecture complies with the converging-diverging paths and ensures the output image size is the same as input image size.  
    
    :return: a keras model for a U-net-like architecture
    :rtype: tf.Keras.model
    
    Parameters
    ----------
    img_shape  : tuple
            input image shape (ny,nx,1)
    n_depth  : int or list
            Option 1: a list of the number of filters in each convolutional layer upstream of each pooling layer. Length must equal number of max pooling layers.  
            
            Option 2: an integer that multiplies the values in this list: [16, 32, ...]. E.g. n_depth = 2 creates [32, 64, ...]  
            
    n_pools  : int
            Number of max pooling layers
    activation : str or tf.Keras.layers.Activation
            name of custom activation or Keras activation layer
    batch_norm : bool
            True to insert BN layer after the convolutional layers
    kern_size  : list or tuple
            kernel size, e.g. (3,3). Provide a list (length = n_pools) of tuples to apply a different kernel size to each block.
    kern_size_upconv  : list or tuple
            kernel size for upconv, e.g. (2,2). Provide a list (length = n_pools) of tuples to apply a different kernel size to each block
    pool_size        : list or tuple
            max pool size, e.g. (2,2). Provide a list (length = n_pools) of tuples to apply a different size to each block
    dropout_level : float or list  
            Option 1:  a list (length = n_pools) of dropout values to apply separately in each block  
            
            Option 2:  a float (0..1) that multiples the values in this list: [0.1, 0.1, 0.2, 0.2, 0.3]  
            
    loss          : str
            The loss function of your choice. The following are implemented:  
            'weighted_crossentropy', 'focal_loss', 'binary_crossentropy'  
    stdinput     : bool
            If True, the input image will be normalized into [0,1]
    """

    
    # Check inputs
    if (img_shape[0] % (2**n_pools) != 0) | (img_shape[1] % (2**n_pools) != 0):
        raise ValueError("Image shape must be divisible by 2^n_pools.")
    
#    n_filters = np.asarray([16, 32, 64, 128, 256])*n_depth 
    n_filters = _expand_inputs(n_depth, n_pools+1, 'n_depth')
    dropouts = _expand_inputs(dropout_level, n_pools+1, 'dropouts')
    np.asarray([0.1, 0.1, 0.2, 0.2, 0.3])*dropout_level 
    
#    n_filters = n_filters[:n_pools+1]
    dropouts = dropouts[:n_pools+1]
    
    if n_pools not in range(2,5):
        raise ValueError("This implementation allows number of pooling layers to be 2, 3 or 4.")
    
    pool_size = _expand_inputs(pool_size, n_pools, "pool_size")
    kern_size = _expand_inputs(kern_size, n_pools + 1, "kern_size")
    kern_size_upconv = _expand_inputs(kern_size_upconv, n_pools, "kern_size_upconv")

    
    inp = L.Input(img_shape)
    
    if stdinput:
        standardizer = L.Lambda(losses.standardize)
        stdinp = standardizer(inp)
    else:
        stdinp = inp
    
    convs, pools = [], []
    
    for ii in range(n_pools):
        
        if ii == 0:
            tensor_in = stdinp
        else:
            tensor_in = pools[ii-1]
        
        conv, pool = pool_layer(tensor_in, n_filters[ii], pool_size[ii], dropout = dropouts[ii], activation = activation, batch_norm = batch_norm, kern_size = kern_size[ii])
        convs.append(conv)
        pools.append(pool)
    
    bottleneck = conv_layer(pools[-1], n_filters[-1], dropout = dropouts[-1], activation = activation, batch_norm = batch_norm, kern_size = kern_size[ii])
    
    
    for ii in range(n_pools-1, -1, -1):
        
        if ii == n_pools - 1:
            tensor = bottleneck
        
        tensor = upconv_layer(tensor, convs[ii], activation = activation, batch_norm = batch_norm, kern_size = kern_size_upconv[ii], strides = pool_size[ii])
        tensor = conv_layer(tensor, n_filters[ii], dropout = dropouts[ii], activation = activation, batch_norm = batch_norm, kern_size = kern_size[ii])
        
    
    
    out = L.Conv2D(1, (1,1), activation =  'sigmoid') (tensor)
    
    model = keras.models.Model(inputs = inp, outputs = out)
    adam = keras.optimizers.Adam(lr=0.0001, beta_1=0.99, beta_2=0.999, epsilon=1e-8, decay=0.0, amsgrad=False)
    #model.compile(optimizer=adam, loss='binary_crossentropy') # Define the optimizer as adamax GD and loss function as mean squared error
    
    if loss == 'weighted_crossentropy':
        loss = losses.weighted_crossentropy
    elif loss == 'my_binary_crossentropy':
        loss = losses.my_binary_crossentropy
    elif loss == 'focal_loss':
        loss = losses.focal_loss
    elif loss == 'binary_crossentropy':
        loss = loss
    else:
        raise ValueError("Loss function not recognized...")
    
    model.compile(optimizer=adam, loss=loss, metrics = [losses.IoU, losses.acc_zeros, losses.acc_ones])
    
    return model




































#def build_Unet_flex(img_shape, n_depth = 1, n_pools = 4, activation = 'lrelu', \
#                     batch_norm = True, kern_size = (3,3), kern_size_upconv = (2,2), pool_size = (2,2), dropout_level = 1.0,\
#                     loss = 'binary_crossentropy', stdinput = False):
#    
#    n_filters = np.asarray([16, 32, 64, 128, 256])*n_depth
#    dropouts = np.asarray([0.1, 0.1, 0.2, 0.2, 0.3])*dropout_level
#    
#    n_filters = n_filters[:n_pools+1]
#    dropouts = dropouts[:n_pools+1]
#    
#    if n_pools not in range(2,5):
#        raise ValueError("This implementation allows number of pooling layers to be 2, 3 or 4.")
#    
#    inp = L.Input(img_shape)
#    
#    if stdinput:
#        standardizer = L.Lambda(losses.standardize)
#        stdinp = standardizer(inp)
#    else:
#        stdinp = inp
#    
#    convs, pools = [], []
#    
#    for ii in range(n_pools):
#        
#        if ii == 0:
#            tensor_in = stdinp
#        else:
#            tensor_in = pools[ii-1]
#        
#        conv, pool = pool_layer(tensor_in, n_filters[ii], pool_size, dropout = dropouts[ii], activation = activation, batch_norm = batch_norm, kern_size = kern_size)
#        convs.append(conv)
#        pools.append(pool)
#    
#    bottleneck = conv_layer(pools[-1], n_filters[-1], dropout = dropouts[-1], activation = activation, batch_norm = batch_norm, kern_size = kern_size)
#    
#    
#    for ii in range(n_pools-1, -1, -1):
#        
#        if ii == n_pools - 1:
#            tensor = bottleneck
#        
#        tensor = upconv_layer(tensor, convs[ii], activation = activation, batch_norm = batch_norm, kern_size = kern_size_upconv, strides = pool_size)
#        tensor = conv_layer(tensor, n_filters[ii], dropout = dropouts[ii], activation = activation, batch_norm = batch_norm, kern_size = kern_size)
#        
#    
#    
#    out = L.Conv2D(1, (1,1), activation =  'sigmoid') (tensor)
#    
#    model = keras.models.Model(inputs = inp, outputs = out)
#    adam = keras.optimizers.Adam(lr=0.0001, beta_1=0.99, beta_2=0.999, epsilon=1e-8, decay=0.0, amsgrad=False)
#    #model.compile(optimizer=adam, loss='binary_crossentropy') # Define the optimizer as adamax GD and loss function as mean squared error
#    
#    if loss == 'weighted_crossentropy':
#        loss = losses.weighted_crossentropy
#    elif loss == 'my_binary_crossentropy':
#        loss = losses.my_binary_crossentropy
#    elif loss == 'focal_loss':
#        loss = losses.focal_loss
#    elif loss == 'binary_crossentropy':
#        loss = loss
#    else:
#        raise ValueError("Loss function not recognized...")
#    
#    model.compile(optimizer=adam, loss=loss, metrics = [losses.IoU, losses.acc_zeros, losses.acc_ones])
#    
#    return model
#
#
#



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 17 13:34:59 2019

@author: atekawade
"""

import numpy as np
from scipy.ndimage import affine_transform





def apply_Affine(vol, M, T = None, order = 1):
    
    # vol    : 3D numpy array
    # M       : 3x3 array as the transform matrix
    # T      : 1x3 array as the translation matrix. None if no translation
    
    return affine_transform(vol, M, T if T is not None else np.zeros(3), order = order)

def resize_volume(vol, new_shape, order = None):

    # vol        : 3D numpy array
    # new_shape  : new shape of array
    # order      : order for spline interpolation
    old_shape = vol.shape
    s = np.asarray([float(old_shape[i])/float(new_shape[i]) for i in range(3)])
    
    M = np.identity(3)*s
    return affine_transform(vol, M, np.zeros(3), order = order, output_shape = new_shape)


def get_patches(vol, patch_size = None, steps = None):
    
    stepz, stepy, stepx = steps
    mz, my, mx = vol.shape
    pz, py, px = patch_size[0], patch_size[1], patch_size[2]
    nx, ny, nz = int(np.ceil(mx/px)), int(np.ceil(my/py)), int(np.ceil(mz/pz))
    
    vol = np.asarray([vol[ii*stepz:ii*stepz+pz] for ii in range(nz)])
    vol = np.asarray([[vol[jj,:,ii*stepy:ii*stepy+py] for ii in range(ny)] for jj in range(vol.shape[0])])
    vol = np.asarray([[[vol[jj,kk,:,:,ii*stepx:ii*stepx+px] for ii in range(nx)] for kk in range(vol.shape[1])] for jj in range(vol.shape[0])])
    return vol


def get_stepsize(vol_shape, patch_size):
    
    # Find optimum number of patches to cover full image
    mz, my, mx = vol_shape
    pz, py, px = patch_size
    nx, ny, nz = int(np.ceil(mx/px)), int(np.ceil(my/py)), int(np.ceil(mz/pz))
    stepx = (mx-px) // (nx-1) if mx != px else 0
    stepy = (my-py) // (ny-1) if my != py else 0
    stepz = (mz-pz) // (nz-1) if mz != pz else 0
    
    return (stepz, stepy, stepx)



def recon_patches(vol, vol_shape = None, steps = None):
    
    if vol.ndim != 6:
        raise ValueError("Input must be 6D array.")
    
    nz, ny, nx, pz, py, px = vol.shape
    stepz, stepy, stepx = steps
    
    
    new_vol = np.zeros((vol_shape), dtype = vol.dtype)
    
    for kk in range(nz):
        for ii in range(ny):
            for jj in range(nx):
                new_vol[kk*stepz:kk*stepz+pz, ii*stepy:ii*stepy+py, jj*stepx:jj*stepx+px] = vol[kk,ii,jj]
        
    return new_vol










#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Oct 10 10:56:50 2018
@author: atekawade
"""

from setuptools import setup, find_packages

setup(
    # Needed to silence warnings (and to be a worthwhile package)
    name='ct_segnet',
    url='https://github.com/aniketkt/CTSegNet',
    author='Aniket Tekawade',
    author_email='atekawade@anl.gov',
    # Needed to actually package something
    packages= ['ct_segnet', 'ct_segnet.data_utils', 'ct_segnet.model_utils'],
    # Needed for dependencies
    install_requires=['numpy', 'scikit-image', 'matplotlib', 'opencv-python', 'pandas', 'tifffile', 'scipy','scikit-learn','ConfigArgParse', 'tqdm', 'ipython', 'h5py'],
    version=open('VERSION').read().strip(),
    license='BSD',
    description='Automated 3D segmentation powered by 2D convolutional neural networks',
    # We will also need a readme eventually (there will be a warning)
    #long_description=open('README.md').read(),
)





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May 20 17:42:56 2019

@author: atekawade
"""

import numpy as np
from multiprocessing import cpu_count
import matplotlib as mpl
from ct_segnet.data_utils.data_io import Parallelize

import matplotlib.pyplot as plt
from sklearn.mixture import GaussianMixture as GMM
from ct_segnet import stats

def _normalize(img, amin = 0.0, amax = 1.0):
    alow = img.min()
    ahigh = img.max()
    
    return np.nan_to_num(amin + (amax-amin)*(img - alow)/(ahigh - alow))

def run_augmenter(x_imgs, y_imgs, to_do = ["rotate", "flip", "gaussian noise", "invert intensity"], nprocs = 1, min_SNR = 2.0, inplace = True):
    
    """
    Parameters
    ----------
    x_imgs : np.array
        input images (n_batch, ny, ny), list of images or 3D numpy array
        
    y_imgs : np.array
        segmented images (n_batch, ny, ny), list of images or 3D numpy array
        
    to_do  : list
        list of strings from "rotate", "flip", "gaussian noise", "invert intensity"
        
    min_SNR : float
        minimum allowable signal-to-noise ratio (SNR)
        
    inplace : bool
        augmentation of data in-place. This is faster and avoids copying the array but could lead to memory-leaks.
        
    """
    if to_do is None: return x_imgs, y_imgs
    
    f = augment_data
    
    if not inplace:
        XY = [(x_imgs[ii], y_imgs[ii]) for ii in range(x_imgs.shape[0])]
        del x_imgs
        del y_imgs

        XY = np.asarray(Parallelize(XY, f, to_do = to_do, min_SNR = min_SNR, procs = nprocs))

        x_imgs = XY[:,0,:,:]
        y_imgs = XY[:,1,:,:]

        return x_imgs, y_imgs
    
    else:
#         print("This loop")
        for i in range(x_imgs.shape[0]):
            x_imgs[i], y_imgs[i] = augment_data(x_imgs[i], y_imgs[i], min_SNR = min_SNR, to_do = to_do)
        return x_imgs, y_imgs

def augment_data(x_img, y_img, to_do = None, min_SNR = 2.0):

    """
    Parameters
    ----------
    x_img : np.array  
        input image (2D)  
        
    y_img : np.array  
        segmentation map (2D)  
    
    Returns
    -------
    tuple
        image pair (input image, segmentation map)
    
    """
    if to_do == None:
        return x_img, y_img
    
    if "gaussian noise" in to_do:
        x_img, y_img = gaussian_noise(x_img, y_img, min_SNR = min_SNR)
    if "rotate" in to_do:
        x_img, y_img = random_rotate(x_img, y_img)
    if "flip" in to_do:
        x_img, y_img = random_flip(x_img, y_img)
    if "intensity flip" in to_do:
        x_img, y_img = invert_intensity(x_img, y_img)
    if "invert intensity" in to_do:
        x_img, y_img = invert_intensity(x_img, y_img)
    
    return x_img, y_img

def invert_intensity(x_img, y_img):
    """
    Darker pixels are made brighter and vice versa. The labels in the corresponding segmentation map are also inverted.
    
    Parameters
    ----------
    x_img : np.array  
        input image (2D)  
        
    y_img : np.array  
        segmentation map (2D)  
    
    Returns
    -------
    tuple
        image pair (input image, segmentation map)
    
    """
    if np.random.randint(2,size = 1)[0]:
        amin, amax = np.min(x_img), np.max(x_img)
        return _normalize(x_img, amin = amax, amax = amin), y_img^1
    else:
        return x_img, y_img

    
    
    
def gaussian_noise(x_img, y_img, min_SNR = 2.0):
    """
    The SNR is first measured in the original image, then normally distributed noise is added such that the SNR does not drop below min_SNR.
    
    Parameters
    ----------
    x_img : np.array  
        input image (2D)  
        
    y_img : np.array  
        segmentation map (2D)  
    
    Returns
    -------
    tuple
        image pair (input image, segmentation map)
    
    """
    SNR0 = stats.calc_SNR(x_img, y_img)
    if min_SNR >= SNR0:
        return x_img, y_img
    else:
        SNR1 = np.random.uniform(min_SNR, SNR0)
        sigma_add = np.sqrt((1/SNR1**2 - 1/SNR0**2))
        return x_img + np.random.normal(0, sigma_add, x_img.shape), y_img

    
def random_flip(x_img, y_img):
    """
    The image and the corresponding seg. map are randomly flipped vertically, horizontally, or both.
    
    Parameters
    ----------
    x_img : np.array  
        input image (2D)  
        
    y_img : np.array  
        segmentation map (2D)  
    
    Returns
    -------
    tuple
        image pair (input image, segmentation map)
    
    """
    
    flags = np.random.randint(2,size=2)
    
    if flags[0] == 1:
        x_img, y_img = np.flip(x_img, axis = 0), np.flip(y_img, axis = 0)
    if flags[1] == 1:
        x_img, y_img = np.flip(x_img, axis = 1), np.flip(y_img, axis = 1)
        
    return x_img, y_img

        
def random_rotate(x_img, y_img):
    """
    The image and the corresponding seg. map are randomly rotated by 90, 180, 270 degrees.
    
    Parameters
    ----------
    x_img : np.array  
        input image (2D)  
        
    y_img : np.array  
        segmentation map (2D)  
    
    Returns
    -------
    tuple
        image pair (input image, segmentation map)
    
    """
    
    nrot = np.random.randint(4, size =1)[0]
    
    return np.rot90(x_img, k = nrot), np.rot90(y_img, k = nrot)
    


def remove_blanks(x_train, y_train, cutoff = 0.3, return_idx = False):

    ystd = np.std(y_train, axis = (1,2))
#     fig, ax = plt.subplots(1,1)
#     ax.plot(np.sort(ystd))
#     ax.plot(np.full(np.shape(ystd), np.max(ystd)*cutoff), label = 'cutoff')
#     ax.legend()
#     plt.savefig("data_variability_total.png")
    
    if cutoff > 0.0:
        idx = np.where(ystd > np.max(ystd)*cutoff)
        y_train = y_train[idx]
        x_train = x_train[idx]

    if return_idx:
        return x_train, y_train, idx
    else:
        return x_train, y_train
    

def _radial_mask(size, rad, inverted = False):
    
    x = np.arange(-size//2, size//2, 1)
    y = np.copy(x)
    xx, yy = np.meshgrid(x, y)
    zz = np.sqrt(xx**2 + yy**2)
    
    if inverted == False:
        bool_mask = zz < rad
    else:
        bool_mask = zz > rad
    
    mask = np.zeros(bool_mask.shape)
    mask[bool_mask] = 1.0
    
    return mask

def apply_circularmask(img, crop_size = None):
    
    if img.ndim == 3:
        ix, iy = 1,2
    elif img.ndim == 2:
        ix, iy = 0,1
    
    if img.shape[ix] != img.shape[iy]:
            raise ValueError("Image must have a square shape.")
    else:
        size  = img.shape[ix]
        
    if crop_size is None:
        crop_size = int(size*0.05)
    
    mask = _radial_mask(size, size//2 - crop_size)
    
    if img.ndim == 3:
        mask = np.asarray([mask for ii in range(img.shape[0])])
        
    img = img*mask
    
    return img
    

def _add_noise(x_img, degree = 2.0, dist = None):
    
    if dist is None:
        raise ValueError("Required keyword argument dist not provided")
    
    fac = max(0, np.random.normal(degree/2,  degree/2))
    
    gauss = np.random.normal(0, dist*fac, x_img.shape)
    return x_img + gauss


def run_add_noise(X, degree = 2.0, max_points = 1e8):

    n_imgs = int(max_points/(np.prod(X.shape[1:])))
    idx = np.random.choice(X.shape[0], n_imgs)
    data_in = X[idx].copy().reshape(-1,1)    
    print("\tEstimating current noise levels...")
    plt.figure()
    whatever = plt.hist(data_in.reshape(-1), bins = 700)
#    plt.savefig('histogram_addnoise.png')
    plt.close()
    model = GMM(2).fit(data_in.reshape(-1,1)) # model size is 2 because binary segmentation
    dist = 0.5*np.abs(model.means_[1][0] - model.means_[0][0])
    print("\tCorrupting images to degree %2.f"%degree)
    
    X = Parallelize(X, _add_noise, degree = degree, dist = dist)
    
    return np.asarray(X)


def range_shift(x_img, y_img):
    """
    Not implemented!  
    
    Stretch the contrast range in the image
    
    Parameters
    ----------
    x_img : np.array  
        input image (2D)  
        
    y_img : np.array  
        segmentation map (2D)  
    
    Returns
    -------
    tuple
        image pair (input image, segmentation map)
    
    """
    
#     amin, amax = -1, -1
#     while ((amin > 1.0) or (amin < 0.0)): 
#         amin = np.random.normal(0.5, 0.25, 1)[0]
    
#     while ((amax > 1.0) or (amax < 0.0) or np.abs(amax - amin) < 0.25):
#         amax = np.random.normal(0.5, 0.25, 1)[0]
        
#     x_img = _normalize(x_img, amin = amin, amax = amax)
    
#     if amax > amin:
#         amax, amin = 1, 0
#     else:
#         amax, amin = 0, 1
#     y_img = _normalize(y_img, amin = amin, amax = amax)

    raise ValueError("Not implemented! Consider using gaussian_noise instead.")
    return x_img, y_img 






















































    
    
    
    
    